
function [basis,v_basis,index_basis] = fundamental_cycle_basis(A,weights,cam)
%
% [basis,v_basis,index_basis] = fundamental_cycle_basis(A,weights,cam)
% This function computes a Fundamental Cycle Basis of an undirected graph,
% i.e. a basis obtained from the cycles induced by a spanning tree.
% It is assumed that the underlying graph is connected.
% The spanning tree is computed as a maximum-weight spanning tree with a
% given root by Kruskal's algorithm, so as to allow the generation of
% multiple random spanning trees. The time complexity of Kruskal's
% algorithm is O(nE+x*log(nV)) where x is the number of edges no longer
% than the longest edge in the spanning tree.
%
% Authors: Federica Arrigoni, Beatrice Rossi, Andrea Fusiello, 2015
%
% INPUT:
% A is the nV x nV (symmetric) adjacency matrix associated to a connected
% graph: A(i,j)=1 if the edge (i,j) appears in the graph, A(i,j) = 0
% otherwise.
%
% [optional] weights are the weights of the edges in the graph used to
% compute a maximum-weight spanning tree. If weights=[] then random
% weights are chosen.
%
% [optional] cam is the root of the spanning tree. If cam=[] then a random
% index between 1 and nV is chosen.
%
% OUTPUT:
% basis is a 1 x nE-nV+1 cell which contains a minimum cycle basis, i.e.
% basis{i} contains the i-th cycle of the basis stored as a closed path,
% e.g. the cycle {(l,j), (j,k), (k,l)} is stored as [l,j,k,l].
%
% v_basis is a nE-nV+1 x nE matrix which contains the cycle basis for the
% graph viewed as a vector space, i.e. v_basis(i,:) is the signed indicator
% vector associated to the i-th cycle. 
%
% index_basis is a 1 x nE-nV+1 cell which contains the nonzero indices of
% each cycle in the basis, i.e. index_basis{i} contains the nonzero indices
% of the edges of the i-th cycle viewed as a vector in GF(2)^nE.
% To explicitly build this (unsigned) vector, use the command
% sparse(index_basis{i},1,1,nE,1). 
%


% check if A is connected
[~,comp]=components(sparse(A));
if length(comp)>1 % if there are multiple connected components
    error('The graph is not connected!')
end

% Consider the lower-triangular adjacency matrix, i.e. the edges (i,j) and
% (j,i) are considered as a single edge.
A = sparse(tril(A,-1));
[I,J]=find(A);

nV=size(A,1); % number of vertices/nodes of the graph
nE=nnz(A); % number of edges

fprintf('\nDimension of the basis: %d\n',nE-nV+1)


% Initialize the basis and all the parameters
basis=cell(1,nE-nV+1);
v_basis=zeros(nE-nV+1,nE);
index_basis=cell(1,nE-nV+1);
if isempty(cam)
    cam=randi(nV); % random root
end
if isempty(weights)
    weights=rand(nE,1); % weights to maximize
end
treeweights=max(weights)-weights+1; % weights to minimize

fprintf('Computing a spanning tree...\n')

% Compute a spanning tree by Kruskal's algorithm (the input must be a
% symmetric matrix).
options.root=cam;
T = kruskal_mst( sparse([I;J],[J;I],[treeweights;treeweights],nV,nV,2*nE),options);
T(T~=0)=1; % remove weights


% Compute the insiemistic difference between A and T.
[I,J]=find(tril(A-T,-1));

fprintf('Computing shortest paths...\n')

% Compute the predecessor and distance matrices using Floyd's algorithm. 
% Paths(x,y) is the node preceding y on the path from x to y.
% D(x,y) indicates the shortest path distance between vertex x and y.
[D,Paths] = floyd_warshall_all_sp(T);
D=floor(D);

for c=1:length(I)
    
    % Find the shortest path in T between I(c) and J(c)
    
    path=zeros(1,D(I(c),J(c))+1);
    j=J(c);
    for k=1:D(I(c),J(c))+1
        path(k)=j;
        j=Paths(I(c),j);
    end
    
    % Compute the cycle generated by T and the edge I(c),J(c)
    % cycle= [path(1), path(2), ..., path(end), path(1)]
    cycle=[path path(1)];
    [v_cycle,v_index]=cycle2vector(cycle,A,nE); % vector representation
    
    basis{c}=cycle;
    v_basis(c,:)=v_cycle;
    index_basis{c}=v_index;
    
end


fprintf('Fundamental cycle basis computed!\n')


end


